#include <stdlib.h>
#include <math.h>
#include "parser.h"
#include "stack.h"

#define OPP_END 6
#define MUL_NUM 2

static char *precedence_table[] = {
  // +-*/()E   E == End of file/stack
    "RRSSSRR", // +
    "RRSSSRR", // -
    "RRRRSRR", // *
    "RRRRSRR", // /
    "SSSSSP1", // (
    "2222222", // ) - Is never stacked
    "SSSSS3D"  // E
};

static char *compat_table[] = {
    "1111Y1Y",
    "1111Y1Y",
    "1111Y1Y",
    "1111Y1Y",
    "1111Y2Y", // (
    "YYYYMY3",
    "YYYYMY3"
};

static void parser_insert_mul(struct stack *opers)
{
    struct lexeme *mul;

    mul = malloc(sizeof(struct lexeme));
    mul->type = LEX_OPERATOR;
    mul->operator = '*';
    mul->opernum = MUL_NUM;

    stack_push(opers, mul);
}

static int parser_compat(struct stack *opers, int *old_type, 
			 struct lexeme *lexeme)
{
    int type, retval, c;

    /* Set type. It is OPP_END for operand, opernum is for operators */
    if (lexeme->type == LEX_OPERAND) 
	type = OPP_END;
    else 
	type = lexeme->opernum;

    retval = 0;
    switch(c = compat_table[*old_type][type]) {
    case 'Y':
	retval = 1;
	break;
    case 'M':
	parser_insert_mul(opers);
	retval = 1;
	break;
    case '1':
	printf("Expected operand at %d\n", lexeme->offset);
	break;
    case '2':
	printf("Empty subexpression at %d\n", lexeme->offset);
	break;
    case '3':
	printf("Expected operator at %d\n", lexeme->offset);
	break;
    default:
	printf("Strange.. got %c\n", c);
    }

    *old_type = type;
    
    return retval;
}

static struct lexeme *parser_make_operand(double res)
{
    struct lexeme *opnd;

    opnd = malloc(sizeof(struct lexeme));
    opnd->type = LEX_OPERAND;
    opnd->value = res;

    return opnd;
}

static int parser_perform(struct lexeme *oper, struct stack *opnds)
{
    struct lexeme *opnd1, *opnd2;
    double res = 0;

    opnd2 = stack_pop(opnds);
    opnd1 = stack_pop(opnds);

    if (!opnd1 || !opnd2) {
	puts("Operand stack underflow!");
	return 1;
    }

    switch(oper->operator) {
    case '+':
	res = opnd1->value + opnd2->value;
	break;
    case '-':
	res = opnd1->value - opnd2->value;
	break;
    case '*':
	res = opnd1->value * opnd2->value;
	break;
    case '/':
	res = opnd1->value / opnd2->value;
	break;
    default:
	puts("Fucked up operator!");
	return 1;
    }

    lexer_destroy_lexeme(opnd1);
    lexer_destroy_lexeme(opnd2);

    stack_push(opnds, parser_make_operand(res));
    return 0;
}

static int parser_loop(struct stack *opers, struct stack *opnds, 
		       struct lexeme *lexeme)
{
    struct lexeme *stacktop;
    char action;
    int topnum = OPP_END, lexnum = OPP_END;
    int p_err = 0;

    while(!p_err) {
	stacktop = stack_top(opers);
	if (stacktop) topnum = stacktop->opernum;
	else topnum = OPP_END;
	if (lexeme) lexnum = lexeme->opernum;
	else lexnum = OPP_END;
	action = precedence_table[topnum][lexnum];
	
	switch(action) {	 
	case 'S':
	    stack_push(opers, lexeme);
	    return 0;
	case 'R':
	    stacktop = stack_pop(opers);
	    p_err = parser_perform(stacktop, opnds);
	    lexer_destroy_lexeme(stacktop);
	    break;
	case 'P':
	    lexer_destroy_lexeme(lexeme);
	    lexeme = stack_pop(opers);
	    lexer_destroy_lexeme(lexeme);
	    return 0;
	case 'D':
	    return 'D';
	    
	case '1':
	    printf("Unmatched opening parenthesis at %d\n", stacktop->offset);
	    return 1;
	case '2':
	    printf("Internal error, ) detected on stack!\n");
	    return 2;
	case '3':
	    printf("Unmatched closing parenthesis at %d\n", lexeme->offset);
	    return 3;
	default:
	    puts("Corrupted precedence-table");
	    return -3;
	}
    }
    return -2;
}

double parse_it(struct lexer *lex)
{
    struct stack *opers = stack_init(), *opnds = stack_init();
    struct lexeme *lexeme;
    double result = NAN;
    int p_stat = 0;
    int old_type = 0;
    
    while (!p_stat) {
	lexeme = lexer_next(lex);
	/* Check if this lexeme is compatible with the previous.
	 * eg: 10 + + 5, the second + is not compatible with the first */
	if (lexeme && !parser_compat(opers, &old_type, lexeme)) {
	    stack_destroy(opers);
	    stack_destroy(opnds);
	    return NAN;
	}

	if (lexeme && lexeme->type == LEX_OPERAND)
	    stack_push(opnds, lexeme);
	else
	    p_stat = parser_loop(opers, opnds, lexeme);
    }

    if (!opers->count && opnds->count == 1) {
	lexeme = stack_pop(opnds);
        result = lexeme->value;
        lexer_destroy_lexeme(lexeme);
    }
    stack_destroy(opers);
    stack_destroy(opnds);

    return result;
}
